@@ -1,11 +1,13 @@
 package main
 
 import (
+	"context"
 	"fmt"
 	"log"
 	"os"
 	"os/signal"
 	"strings"
+	"sync/atomic"
 	"syscall"
 	"time"
 
@@ -14,6 +16,8 @@ import (
 	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmctl/prometheus"
 	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmctl/vm"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/buildinfo"
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/protoparser/common"
+	parser "github.com/VictoriaMetrics/VictoriaMetrics/lib/protoparser/native"
 	"github.com/urfave/cli/v2"
 )
 
@@ -23,6 +27,7 @@ func main() {
 		importer *vm.Importer
 	)
 
+	ctx, cancelCtx := context.WithCancel(context.Background())
 	start := time.Now()
 	app := &cli.App{
 		Name:    "vmctl",
@@ -52,6 +57,9 @@ func main() {
 					}
 
 					vmCfg := initConfigVM(c)
+					// disable progress bars since openTSDB implementation
+					// does not use progress bar pool
+					vmCfg.DisableProgressBar = true
 					importer, err := vm.NewImporter(vmCfg)
 					if err != nil {
 						return fmt.Errorf("failed to create VM importer: %s", err)
@@ -92,8 +100,13 @@ func main() {
 						return fmt.Errorf("failed to create VM importer: %s", err)
 					}
 
-					processor := newInfluxProcessor(influxClient, importer,
-						c.Int(influxConcurrency), c.String(influxMeasurementFieldSeparator))
+					processor := newInfluxProcessor(
+						influxClient,
+						importer,
+						c.Int(influxConcurrency),
+						c.String(influxMeasurementFieldSeparator),
+						c.Bool(influxSkipDatabaseLabel),
+						c.Bool(influxPrometheusMode))
 					return processor.run(c.Bool(globalSilent), c.Bool(globalVerbose))
 				},
 			},
@@ -161,7 +174,40 @@ func main() {
 							extraLabels: c.StringSlice(vmExtraLabel),
 						},
 					}
-					return p.run()
+					return p.run(ctx)
+				},
+			},
+			{
+				Name:  "verify-block",
+				Usage: "Verifies exported block with VictoriaMetrics Native format",
+				Flags: []cli.Flag{
+					&cli.BoolFlag{
+						Name:  "gunzip",
+						Usage: "Use GNU zip decompression for exported block",
+						Value: false,
+					},
+				},
+				Action: func(c *cli.Context) error {
+					common.StartUnmarshalWorkers()
+					blockPath := c.Args().First()
+					isBlockGzipped := c.Bool("gunzip")
+					if len(blockPath) == 0 {
+						return cli.Exit("you must provide path for exported data block", 1)
+					}
+					log.Printf("verifying block at path=%q", blockPath)
+					f, err := os.OpenFile(blockPath, os.O_RDONLY, 0600)
+					if err != nil {
+						return cli.Exit(fmt.Errorf("cannot open exported block at path=%q err=%w", blockPath, err), 1)
+					}
+					var blocksCount uint64
+					if err := parser.ParseStream(f, isBlockGzipped, func(block *parser.Block) error {
+						atomic.AddUint64(&blocksCount, 1)
+						return nil
+					}); err != nil {
+						return cli.Exit(fmt.Errorf("cannot parse block at path=%q, blocksCount=%d, err=%w", blockPath, blocksCount, err), 1)
+					}
+					log.Printf("successfully verified block at path=%q, blockCount=%d", blockPath, blocksCount)
+					return nil
 				},
 			},
 		},
@@ -175,11 +221,12 @@ func main() {
 		if importer != nil {
 			importer.Close()
 		}
+		cancelCtx()
 	}()
 
 	err = app.Run(os.Args)
 	if err != nil {
-		log.Println(err)
+		log.Fatalln(err)
 	}
 	log.Printf("Total time: %v", time.Since(start))
 }
@@ -197,5 +244,6 @@ func initConfigVM(c *cli.Context) vm.Config {
 		RoundDigits:        c.Int(vmRoundDigits),
 		ExtraLabels:        c.StringSlice(vmExtraLabel),
 		RateLimit:          c.Int64(vmRateLimit),
+		DisableProgressBar: c.Bool(vmDisableProgressBar),
 	}
 }
