@@ -13,8 +13,10 @@ import (
 	"sync"
 	"time"
 
+	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmctl/barpool"
 	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmctl/limiter"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/decimal"
+	"github.com/dmitryk-dk/pb/v3"
 )
 
 // Config contains list of params to configure
@@ -51,6 +53,8 @@ type Config struct {
 	// RateLimit defines a data transfer speed in bytes per second.
 	// Is applied to each worker (see Concurrency) independently.
 	RateLimit int64
+	// Whether to disable progress bar per VM worker
+	DisableProgressBar bool
 }
 
 // Importer performs insertion of timeseries
@@ -144,15 +148,22 @@ func NewImporter(cfg Config) (*Importer, error) {
 
 	im.wg.Add(int(cfg.Concurrency))
 	for i := 0; i < int(cfg.Concurrency); i++ {
-		go func() {
+		var bar *pb.ProgressBar
+		if !cfg.DisableProgressBar {
+			pbPrefix := fmt.Sprintf(`{{ green "VM worker %d:" }}`, i)
+			bar = barpool.AddWithTemplate(pbPrefix+pbTpl, 0)
+		}
+		go func(bar *pb.ProgressBar) {
 			defer im.wg.Done()
-			im.startWorker(cfg.BatchSize, cfg.SignificantFigures, cfg.RoundDigits)
-		}()
+			im.startWorker(bar, cfg.BatchSize, cfg.SignificantFigures, cfg.RoundDigits)
+		}(bar)
 	}
 	im.ResetStats()
 	return im, nil
 }
 
+const pbTpl = `{{ (cycle . "←" "↖" "↑" "↗" "→" "↘" "↓" "↙" ) }} {{speed . "%s samples/s"}}`
+
 // ImportError is type of error generated
 // in case of unsuccessful import request
 type ImportError struct {
@@ -170,7 +181,17 @@ func (im *Importer) Errors() chan *ImportError { return im.errors }
 
 // Input returns a channel for sending timeseries
 // that need to be imported
-func (im *Importer) Input() chan<- *TimeSeries { return im.input }
+func (im *Importer) Input(ts *TimeSeries) error {
+	select {
+	case im.input <- ts:
+		return nil
+	case err := <-im.errors:
+		if err != nil && err.Err != nil {
+			return err.Err
+		}
+		return fmt.Errorf("process aborted")
+	}
+}
 
 // Close sends signal to all goroutines to exit
 // and waits until they are finished
@@ -182,7 +203,7 @@ func (im *Importer) Close() {
 	})
 }
 
-func (im *Importer) startWorker(batchSize, significantFigures, roundDigits int) {
+func (im *Importer) startWorker(bar *pb.ProgressBar, batchSize, significantFigures, roundDigits int) {
 	var batch []*TimeSeries
 	var dataPoints int
 	var waitForBatch time.Time
@@ -217,6 +238,11 @@ func (im *Importer) startWorker(batchSize, significantFigures, roundDigits int)
 
 			batch = append(batch, ts)
 			dataPoints += len(ts.Values)
+
+			if bar != nil {
+				bar.Add(len(ts.Values))
+			}
+
 			if dataPoints < batchSize {
 				continue
 			}
@@ -232,8 +258,8 @@ func (im *Importer) startWorker(batchSize, significantFigures, roundDigits int)
 				// make a new batch, since old one was referenced as err
 				batch = make([]*TimeSeries, len(batch))
 			}
-			batch = batch[:0]
 			dataPoints = 0
+			batch = batch[:0]
 			waitForBatch = time.Now()
 		}
 	}
