@@ -8,6 +8,7 @@ import (
 	"sort"
 	"strings"
 	"sync"
+	"sync/atomic"
 
 	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmselect/netstorage"
 	"github.com/VictoriaMetrics/VictoriaMetrics/app/vmselect/searchutils"
@@ -16,6 +17,7 @@ import (
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/decimal"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/logger"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/memory"
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/querytracer"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/storage"
 	"github.com/VictoriaMetrics/metrics"
 	"github.com/VictoriaMetrics/metricsql"
@@ -93,11 +95,16 @@ type EvalConfig struct {
 	End   int64
 	Step  int64
 
+	// MaxSeries is the maximum number of time series, which can be scanned by the query.
+	// Zero means 'no limit'
+	MaxSeries int
+
 	// QuotedRemoteAddr contains quoted remote address.
 	QuotedRemoteAddr string
 
 	Deadline searchutils.Deadline
 
+	// Whether the response can be cached.
 	MayCache bool
 
 	// LookbackDelta is analog to `-query.lookback-delta` from Prometheus.
@@ -113,12 +120,13 @@ type EvalConfig struct {
 	timestampsOnce sync.Once
 }
 
-// newEvalConfig returns new EvalConfig copy from src.
-func newEvalConfig(src *EvalConfig) *EvalConfig {
+// copyEvalConfig returns src copy.
+func copyEvalConfig(src *EvalConfig) *EvalConfig {
 	var ec EvalConfig
 	ec.Start = src.Start
 	ec.End = src.End
 	ec.Step = src.Step
+	ec.MaxSeries = src.MaxSeries
 	ec.Deadline = src.Deadline
 	ec.MayCache = src.MayCache
 	ec.LookbackDelta = src.LookbackDelta
@@ -154,6 +162,12 @@ func (ec *EvalConfig) mayCache() bool {
 	return true
 }
 
+func (ec *EvalConfig) timeRangeString() string {
+	start := storage.TimestampToHumanReadableFormat(ec.Start)
+	end := storage.TimestampToHumanReadableFormat(ec.End)
+	return fmt.Sprintf("[%s..%s]", start, end)
+}
+
 func (ec *EvalConfig) getSharedTimestamps() []int64 {
 	ec.timestampsOnce.Do(ec.timestampsInit)
 	return ec.timestamps
@@ -185,19 +199,42 @@ func getTimestamps(start, end, step int64) []int64 {
 	return timestamps
 }
 
-func evalExpr(ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
+func evalExpr(qt *querytracer.Tracer, ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
+	if qt.Enabled() {
+		query := string(e.AppendString(nil))
+		query = bytesutil.LimitStringLen(query, 300)
+		mayCache := ec.mayCache()
+		qt = qt.NewChild("eval: query=%s, timeRange=%s, step=%d, mayCache=%v", query, ec.timeRangeString(), ec.Step, mayCache)
+	}
+	rv, err := evalExprInternal(qt, ec, e)
+	if err != nil {
+		return nil, err
+	}
+	if qt.Enabled() {
+		seriesCount := len(rv)
+		pointsPerSeries := 0
+		if len(rv) > 0 {
+			pointsPerSeries = len(rv[0].Timestamps)
+		}
+		pointsCount := seriesCount * pointsPerSeries
+		qt.Donef("series=%d, points=%d, pointsPerSeries=%d", seriesCount, pointsCount, pointsPerSeries)
+	}
+	return rv, nil
+}
+
+func evalExprInternal(qt *querytracer.Tracer, ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
 	if me, ok := e.(*metricsql.MetricExpr); ok {
 		re := &metricsql.RollupExpr{
 			Expr: me,
 		}
-		rv, err := evalRollupFunc(ec, "default_rollup", rollupDefault, e, re, nil)
+		rv, err := evalRollupFunc(qt, ec, "default_rollup", rollupDefault, e, re, nil)
 		if err != nil {
 			return nil, fmt.Errorf(`cannot evaluate %q: %w`, me.AppendString(nil), err)
 		}
 		return rv, nil
 	}
 	if re, ok := e.(*metricsql.RollupExpr); ok {
-		rv, err := evalRollupFunc(ec, "default_rollup", rollupDefault, e, re, nil)
+		rv, err := evalRollupFunc(qt, ec, "default_rollup", rollupDefault, e, re, nil)
 		if err != nil {
 			return nil, fmt.Errorf(`cannot evaluate %q: %w`, re.AppendString(nil), err)
 		}
@@ -206,26 +243,12 @@ func evalExpr(ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
 	if fe, ok := e.(*metricsql.FuncExpr); ok {
 		nrf := getRollupFunc(fe.Name)
 		if nrf == nil {
-			args, err := evalExprs(ec, fe.Args)
-			if err != nil {
-				return nil, err
-			}
-			tf := getTransformFunc(fe.Name)
-			if tf == nil {
-				return nil, fmt.Errorf(`unknown func %q`, fe.Name)
-			}
-			tfa := &transformFuncArg{
-				ec:   ec,
-				fe:   fe,
-				args: args,
-			}
-			rv, err := tf(tfa)
-			if err != nil {
-				return nil, fmt.Errorf(`cannot evaluate %q: %w`, fe.AppendString(nil), err)
-			}
-			return rv, nil
+			qtChild := qt.NewChild("transform %s()", fe.Name)
+			rv, err := evalTransformFunc(qtChild, ec, fe)
+			qtChild.Donef("series=%d", len(rv))
+			return rv, err
 		}
-		args, re, err := evalRollupFuncArgs(ec, fe)
+		args, re, err := evalRollupFuncArgs(qt, ec, fe)
 		if err != nil {
 			return nil, err
 		}
@@ -233,79 +256,23 @@ func evalExpr(ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
 		if err != nil {
 			return nil, err
 		}
-		rv, err := evalRollupFunc(ec, fe.Name, rf, e, re, nil)
+		rv, err := evalRollupFunc(qt, ec, fe.Name, rf, e, re, nil)
 		if err != nil {
 			return nil, fmt.Errorf(`cannot evaluate %q: %w`, fe.AppendString(nil), err)
 		}
 		return rv, nil
 	}
 	if ae, ok := e.(*metricsql.AggrFuncExpr); ok {
-		if callbacks := getIncrementalAggrFuncCallbacks(ae.Name); callbacks != nil {
-			fe, nrf := tryGetArgRollupFuncWithMetricExpr(ae)
-			if fe != nil {
-				// There is an optimized path for calculating metricsql.AggrFuncExpr over rollupFunc over metricsql.MetricExpr.
-				// The optimized path saves RAM for aggregates over big number of time series.
-				args, re, err := evalRollupFuncArgs(ec, fe)
-				if err != nil {
-					return nil, err
-				}
-				rf, err := nrf(args)
-				if err != nil {
-					return nil, err
-				}
-				iafc := newIncrementalAggrFuncContext(ae, callbacks)
-				return evalRollupFunc(ec, fe.Name, rf, e, re, iafc)
-			}
-		}
-		args, err := evalExprs(ec, ae.Args)
-		if err != nil {
-			return nil, err
-		}
-		af := getAggrFunc(ae.Name)
-		if af == nil {
-			return nil, fmt.Errorf(`unknown func %q`, ae.Name)
-		}
-		afa := &aggrFuncArg{
-			ae:   ae,
-			args: args,
-			ec:   ec,
-		}
-		rv, err := af(afa)
-		if err != nil {
-			return nil, fmt.Errorf(`cannot evaluate %q: %w`, ae.AppendString(nil), err)
-		}
-		return rv, nil
+		qtChild := qt.NewChild("aggregate %s()", ae.Name)
+		rv, err := evalAggrFunc(qtChild, ec, ae)
+		qtChild.Donef("series=%d", len(rv))
+		return rv, err
 	}
 	if be, ok := e.(*metricsql.BinaryOpExpr); ok {
-		bf := getBinaryOpFunc(be.Op)
-		if bf == nil {
-			return nil, fmt.Errorf(`unknown binary op %q`, be.Op)
-		}
-		var err error
-		var tssLeft, tssRight []*timeseries
-		switch strings.ToLower(be.Op) {
-		case "and", "if":
-			// Fetch right-side series at first, since it usually contains
-			// lower number of time series for `and` and `if` operator.
-			// This should produce more specific label filters for the left side of the query.
-			// This, in turn, should reduce the time to select series for the left side of the query.
-			tssRight, tssLeft, err = execBinaryOpArgs(ec, be.Right, be.Left, be)
-		default:
-			tssLeft, tssRight, err = execBinaryOpArgs(ec, be.Left, be.Right, be)
-		}
-		if err != nil {
-			return nil, fmt.Errorf("cannot execute %q: %w", be.AppendString(nil), err)
-		}
-		bfa := &binaryOpFuncArg{
-			be:    be,
-			left:  tssLeft,
-			right: tssRight,
-		}
-		rv, err := bf(bfa)
-		if err != nil {
-			return nil, fmt.Errorf(`cannot evaluate %q: %w`, be.AppendString(nil), err)
-		}
-		return rv, nil
+		qtChild := qt.NewChild("binary op %q", be.Op)
+		rv, err := evalBinaryOp(qtChild, ec, be)
+		qtChild.Donef("series=%d", len(rv))
+		return rv, err
 	}
 	if ne, ok := e.(*metricsql.NumberExpr); ok {
 		rv := evalNumber(ec, ne.N)
@@ -324,7 +291,98 @@ func evalExpr(ec *EvalConfig, e metricsql.Expr) ([]*timeseries, error) {
 	return nil, fmt.Errorf("unexpected expression %q", e.AppendString(nil))
 }
 
-func execBinaryOpArgs(ec *EvalConfig, exprFirst, exprSecond metricsql.Expr, be *metricsql.BinaryOpExpr) ([]*timeseries, []*timeseries, error) {
+func evalTransformFunc(qt *querytracer.Tracer, ec *EvalConfig, fe *metricsql.FuncExpr) ([]*timeseries, error) {
+	tf := getTransformFunc(fe.Name)
+	if tf == nil {
+		return nil, fmt.Errorf(`unknown func %q`, fe.Name)
+	}
+	args, err := evalExprs(qt, ec, fe.Args)
+	if err != nil {
+		return nil, err
+	}
+	tfa := &transformFuncArg{
+		ec:   ec,
+		fe:   fe,
+		args: args,
+	}
+	rv, err := tf(tfa)
+	if err != nil {
+		return nil, fmt.Errorf(`cannot evaluate %q: %w`, fe.AppendString(nil), err)
+	}
+	return rv, nil
+}
+
+func evalAggrFunc(qt *querytracer.Tracer, ec *EvalConfig, ae *metricsql.AggrFuncExpr) ([]*timeseries, error) {
+	if callbacks := getIncrementalAggrFuncCallbacks(ae.Name); callbacks != nil {
+		fe, nrf := tryGetArgRollupFuncWithMetricExpr(ae)
+		if fe != nil {
+			// There is an optimized path for calculating metricsql.AggrFuncExpr over rollupFunc over metricsql.MetricExpr.
+			// The optimized path saves RAM for aggregates over big number of time series.
+			args, re, err := evalRollupFuncArgs(qt, ec, fe)
+			if err != nil {
+				return nil, err
+			}
+			rf, err := nrf(args)
+			if err != nil {
+				return nil, err
+			}
+			iafc := newIncrementalAggrFuncContext(ae, callbacks)
+			return evalRollupFunc(qt, ec, fe.Name, rf, ae, re, iafc)
+		}
+	}
+	args, err := evalExprs(qt, ec, ae.Args)
+	if err != nil {
+		return nil, err
+	}
+	af := getAggrFunc(ae.Name)
+	if af == nil {
+		return nil, fmt.Errorf(`unknown func %q`, ae.Name)
+	}
+	afa := &aggrFuncArg{
+		ae:   ae,
+		args: args,
+		ec:   ec,
+	}
+	rv, err := af(afa)
+	if err != nil {
+		return nil, fmt.Errorf(`cannot evaluate %q: %w`, ae.AppendString(nil), err)
+	}
+	return rv, nil
+}
+
+func evalBinaryOp(qt *querytracer.Tracer, ec *EvalConfig, be *metricsql.BinaryOpExpr) ([]*timeseries, error) {
+	bf := getBinaryOpFunc(be.Op)
+	if bf == nil {
+		return nil, fmt.Errorf(`unknown binary op %q`, be.Op)
+	}
+	var err error
+	var tssLeft, tssRight []*timeseries
+	switch strings.ToLower(be.Op) {
+	case "and", "if":
+		// Fetch right-side series at first, since it usually contains
+		// lower number of time series for `and` and `if` operator.
+		// This should produce more specific label filters for the left side of the query.
+		// This, in turn, should reduce the time to select series for the left side of the query.
+		tssRight, tssLeft, err = execBinaryOpArgs(qt, ec, be.Right, be.Left, be)
+	default:
+		tssLeft, tssRight, err = execBinaryOpArgs(qt, ec, be.Left, be.Right, be)
+	}
+	if err != nil {
+		return nil, fmt.Errorf("cannot execute %q: %w", be.AppendString(nil), err)
+	}
+	bfa := &binaryOpFuncArg{
+		be:    be,
+		left:  tssLeft,
+		right: tssRight,
+	}
+	rv, err := bf(bfa)
+	if err != nil {
+		return nil, fmt.Errorf(`cannot evaluate %q: %w`, be.AppendString(nil), err)
+	}
+	return rv, nil
+}
+
+func execBinaryOpArgs(qt *querytracer.Tracer, ec *EvalConfig, exprFirst, exprSecond metricsql.Expr, be *metricsql.BinaryOpExpr) ([]*timeseries, []*timeseries, error) {
 	// Execute binary operation in the following way:
 	//
 	// 1) execute the exprFirst
@@ -348,7 +406,7 @@ func execBinaryOpArgs(ec *EvalConfig, exprFirst, exprSecond metricsql.Expr, be *
 	//
 	// - Queries, which get additional labels from `info` metrics.
 	//   See https://www.robustperception.io/exposing-the-software-version-to-prometheus
-	tssFirst, err := evalExpr(ec, exprFirst)
+	tssFirst, err := evalExpr(qt, ec, exprFirst)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -361,7 +419,7 @@ func execBinaryOpArgs(ec *EvalConfig, exprFirst, exprSecond metricsql.Expr, be *
 		lfs = metricsql.TrimFiltersByGroupModifier(lfs, be)
 		exprSecond = metricsql.PushdownBinaryOpFilters(exprSecond, lfs)
 	}
-	tssSecond, err := evalExpr(ec, exprSecond)
+	tssSecond, err := evalExpr(qt, ec, exprSecond)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -498,10 +556,10 @@ func tryGetArgRollupFuncWithMetricExpr(ae *metricsql.AggrFuncExpr) (*metricsql.F
 	return nil, nil
 }
 
-func evalExprs(ec *EvalConfig, es []metricsql.Expr) ([][]*timeseries, error) {
+func evalExprs(qt *querytracer.Tracer, ec *EvalConfig, es []metricsql.Expr) ([][]*timeseries, error) {
 	var rvs [][]*timeseries
 	for _, e := range es {
-		rv, err := evalExpr(ec, e)
+		rv, err := evalExpr(qt, ec, e)
 		if err != nil {
 			return nil, err
 		}
@@ -510,7 +568,7 @@ func evalExprs(ec *EvalConfig, es []metricsql.Expr) ([][]*timeseries, error) {
 	return rvs, nil
 }
 
-func evalRollupFuncArgs(ec *EvalConfig, fe *metricsql.FuncExpr) ([]interface{}, *metricsql.RollupExpr, error) {
+func evalRollupFuncArgs(qt *querytracer.Tracer, ec *EvalConfig, fe *metricsql.FuncExpr) ([]interface{}, *metricsql.RollupExpr, error) {
 	var re *metricsql.RollupExpr
 	rollupArgIdx := metricsql.GetRollupArgIdx(fe)
 	if len(fe.Args) <= rollupArgIdx {
@@ -523,7 +581,7 @@ func evalRollupFuncArgs(ec *EvalConfig, fe *metricsql.FuncExpr) ([]interface{},
 			args[i] = re
 			continue
 		}
-		ts, err := evalExpr(ec, arg)
+		ts, err := evalExpr(qt, ec, arg)
 		if err != nil {
 			return nil, nil, fmt.Errorf("cannot evaluate arg #%d for %q: %w", i+1, fe.AppendString(nil), err)
 		}
@@ -563,11 +621,12 @@ func getRollupExprArg(arg metricsql.Expr) *metricsql.RollupExpr {
 // expr may contain:
 // - rollupFunc(m) if iafc is nil
 // - aggrFunc(rollupFunc(m)) if iafc isn't nil
-func evalRollupFunc(ec *EvalConfig, funcName string, rf rollupFunc, expr metricsql.Expr, re *metricsql.RollupExpr, iafc *incrementalAggrFuncContext) ([]*timeseries, error) {
+func evalRollupFunc(qt *querytracer.Tracer, ec *EvalConfig, funcName string, rf rollupFunc, expr metricsql.Expr,
+	re *metricsql.RollupExpr, iafc *incrementalAggrFuncContext) ([]*timeseries, error) {
 	if re.At == nil {
-		return evalRollupFuncWithoutAt(ec, funcName, rf, expr, re, iafc)
+		return evalRollupFuncWithoutAt(qt, ec, funcName, rf, expr, re, iafc)
 	}
-	tssAt, err := evalExpr(ec, re.At)
+	tssAt, err := evalExpr(qt, ec, re.At)
 	if err != nil {
 		return nil, fmt.Errorf("cannot evaluate `@` modifier: %w", err)
 	}
@@ -575,10 +634,10 @@ func evalRollupFunc(ec *EvalConfig, funcName string, rf rollupFunc, expr metrics
 		return nil, fmt.Errorf("`@` modifier must return a single series; it returns %d series instead", len(tssAt))
 	}
 	atTimestamp := int64(tssAt[0].Values[0] * 1000)
-	ecNew := newEvalConfig(ec)
+	ecNew := copyEvalConfig(ec)
 	ecNew.Start = atTimestamp
 	ecNew.End = atTimestamp
-	tss, err := evalRollupFuncWithoutAt(ecNew, funcName, rf, expr, re, iafc)
+	tss, err := evalRollupFuncWithoutAt(qt, ecNew, funcName, rf, expr, re, iafc)
 	if err != nil {
 		return nil, err
 	}
@@ -596,13 +655,14 @@ func evalRollupFunc(ec *EvalConfig, funcName string, rf rollupFunc, expr metrics
 	return tss, nil
 }
 
-func evalRollupFuncWithoutAt(ec *EvalConfig, funcName string, rf rollupFunc, expr metricsql.Expr, re *metricsql.RollupExpr, iafc *incrementalAggrFuncContext) ([]*timeseries, error) {
+func evalRollupFuncWithoutAt(qt *querytracer.Tracer, ec *EvalConfig, funcName string, rf rollupFunc,
+	expr metricsql.Expr, re *metricsql.RollupExpr, iafc *incrementalAggrFuncContext) ([]*timeseries, error) {
 	funcName = strings.ToLower(funcName)
 	ecNew := ec
 	var offset int64
 	if re.Offset != nil {
 		offset = re.Offset.Duration(ec.Step)
-		ecNew = newEvalConfig(ecNew)
+		ecNew = copyEvalConfig(ecNew)
 		ecNew.Start -= offset
 		ecNew.End -= offset
 		// There is no need in calling AdjustStartEnd() on ecNew if ecNew.MayCache is set to true,
@@ -615,7 +675,7 @@ func evalRollupFuncWithoutAt(ec *EvalConfig, funcName string, rf rollupFunc, exp
 		// in order to obtain expected OHLC results.
 		// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/309#issuecomment-582113462
 		step := ecNew.Step
-		ecNew = newEvalConfig(ecNew)
+		ecNew = copyEvalConfig(ecNew)
 		ecNew.Start += step
 		ecNew.End += step
 		offset -= step
@@ -623,12 +683,12 @@ func evalRollupFuncWithoutAt(ec *EvalConfig, funcName string, rf rollupFunc, exp
 	var rvs []*timeseries
 	var err error
 	if me, ok := re.Expr.(*metricsql.MetricExpr); ok {
-		rvs, err = evalRollupFuncWithMetricExpr(ecNew, funcName, rf, expr, me, iafc, re.Window)
+		rvs, err = evalRollupFuncWithMetricExpr(qt, ecNew, funcName, rf, expr, me, iafc, re.Window)
 	} else {
 		if iafc != nil {
 			logger.Panicf("BUG: iafc must be nil for rollup %q over subquery %q", funcName, re.AppendString(nil))
 		}
-		rvs, err = evalRollupFuncWithSubquery(ecNew, funcName, rf, expr, re)
+		rvs, err = evalRollupFuncWithSubquery(qt, ecNew, funcName, rf, expr, re)
 	}
 	if err != nil {
 		return nil, err
@@ -671,15 +731,17 @@ func aggregateAbsentOverTime(ec *EvalConfig, expr metricsql.Expr, tss []*timeser
 	return rvs
 }
 
-func evalRollupFuncWithSubquery(ec *EvalConfig, funcName string, rf rollupFunc, expr metricsql.Expr, re *metricsql.RollupExpr) ([]*timeseries, error) {
+func evalRollupFuncWithSubquery(qt *querytracer.Tracer, ec *EvalConfig, funcName string, rf rollupFunc, expr metricsql.Expr, re *metricsql.RollupExpr) ([]*timeseries, error) {
 	// TODO: determine whether to use rollupResultCacheV here.
+	qt = qt.NewChild("subquery")
+	defer qt.Done()
 	step := re.Step.Duration(ec.Step)
 	if step == 0 {
 		step = ec.Step
 	}
 	window := re.Window.Duration(ec.Step)
 
-	ecSQ := newEvalConfig(ec)
+	ecSQ := copyEvalConfig(ec)
 	ecSQ.Start -= window + maxSilenceInterval + step
 	ecSQ.End += step
 	ecSQ.Step = step
@@ -688,7 +750,7 @@ func evalRollupFuncWithSubquery(ec *EvalConfig, funcName string, rf rollupFunc,
 	}
 	// unconditionally align start and end args to step for subquery as Prometheus does.
 	ecSQ.Start, ecSQ.End = alignStartEnd(ecSQ.Start, ecSQ.End, ecSQ.Step)
-	tssSQ, err := evalExpr(ecSQ, re.Expr)
+	tssSQ, err := evalExpr(qt, ecSQ, re.Expr)
 	if err != nil {
 		return nil, err
 	}
@@ -702,29 +764,36 @@ func evalRollupFuncWithSubquery(ec *EvalConfig, funcName string, rf rollupFunc,
 	}
 	tss := make([]*timeseries, 0, len(tssSQ)*len(rcs))
 	var tssLock sync.Mutex
+	var samplesScannedTotal uint64
 	keepMetricNames := getKeepMetricNames(expr)
 	doParallel(tssSQ, func(tsSQ *timeseries, values []float64, timestamps []int64) ([]float64, []int64) {
 		values, timestamps = removeNanValues(values[:0], timestamps[:0], tsSQ.Values, tsSQ.Timestamps)
 		preFunc(values, timestamps)
 		for _, rc := range rcs {
 			if tsm := newTimeseriesMap(funcName, keepMetricNames, sharedTimestamps, &tsSQ.MetricName); tsm != nil {
-				rc.DoTimeseriesMap(tsm, values, timestamps)
+				samplesScanned := rc.DoTimeseriesMap(tsm, values, timestamps)
+				atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 				tssLock.Lock()
 				tss = tsm.AppendTimeseriesTo(tss)
 				tssLock.Unlock()
 				continue
 			}
 			var ts timeseries
-			doRollupForTimeseries(funcName, keepMetricNames, rc, &ts, &tsSQ.MetricName, values, timestamps, sharedTimestamps)
+			samplesScanned := doRollupForTimeseries(funcName, keepMetricNames, rc, &ts, &tsSQ.MetricName, values, timestamps, sharedTimestamps)
+			atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 			tssLock.Lock()
 			tss = append(tss, &ts)
 			tssLock.Unlock()
 		}
 		return values, timestamps
 	})
+	rowsScannedPerQuery.Update(float64(samplesScannedTotal))
+	qt.Printf("rollup %s() over %d series returned by subquery: series=%d, samplesScanned=%d", funcName, len(tssSQ), len(tss), samplesScannedTotal)
 	return tss, nil
 }
 
+var rowsScannedPerQuery = metrics.NewHistogram(`vm_rows_scanned_per_query`)
+
 func getKeepMetricNames(expr metricsql.Expr) bool {
 	if ae, ok := expr.(*metricsql.AggrFuncExpr); ok {
 		// Extract rollupFunc(...) from aggrFunc(rollupFunc(...)).
@@ -797,15 +866,22 @@ var (
 	rollupResultCacheMiss        = metrics.NewCounter(`vm_rollup_result_cache_miss_total`)
 )
 
-func evalRollupFuncWithMetricExpr(ec *EvalConfig, funcName string, rf rollupFunc,
+func evalRollupFuncWithMetricExpr(qt *querytracer.Tracer, ec *EvalConfig, funcName string, rf rollupFunc,
 	expr metricsql.Expr, me *metricsql.MetricExpr, iafc *incrementalAggrFuncContext, windowExpr *metricsql.DurationExpr) ([]*timeseries, error) {
+	var rollupMemorySize int64
+	window := windowExpr.Duration(ec.Step)
+	if qt.Enabled() {
+		qt = qt.NewChild("rollup %s(): timeRange=%s, step=%d, window=%d", funcName, ec.timeRangeString(), ec.Step, window)
+		defer func() {
+			qt.Donef("neededMemoryBytes=%d", rollupMemorySize)
+		}()
+	}
 	if me.IsEmpty() {
 		return evalNumber(ec, nan), nil
 	}
-	window := windowExpr.Duration(ec.Step)
 
 	// Search for partial results in cache.
-	tssCached, start := rollupResultCacheV.Get(ec, expr, window)
+	tssCached, start := rollupResultCacheV.Get(qt, ec, expr, window)
 	if start > ec.End {
 		// The result is fully cached.
 		rollupResultCacheFullHits.Inc()
@@ -834,8 +910,8 @@ func evalRollupFuncWithMetricExpr(ec *EvalConfig, funcName string, rf rollupFunc
 	} else {
 		minTimestamp -= ec.Step
 	}
-	sq := storage.NewSearchQuery(minTimestamp, ec.End, tfss)
-	rss, err := netstorage.ProcessSearchQuery(sq, true, ec.Deadline)
+	sq := storage.NewSearchQuery(minTimestamp, ec.End, tfss, ec.MaxSeries)
+	rss, err := netstorage.ProcessSearchQuery(qt, sq, ec.Deadline)
 	if err != nil {
 		return nil, err
 	}
@@ -869,7 +945,7 @@ func evalRollupFuncWithMetricExpr(ec *EvalConfig, funcName string, rf rollupFunc
 		}
 	}
 	rollupPoints := mulNoOverflow(pointsPerTimeseries, int64(timeseriesLen*len(rcs)))
-	rollupMemorySize := mulNoOverflow(rollupPoints, 16)
+	rollupMemorySize = mulNoOverflow(rollupPoints, 16)
 	rml := getRollupMemoryLimiter()
 	if !rml.Get(uint64(rollupMemorySize)) {
 		rss.Cancel()
@@ -886,15 +962,15 @@ func evalRollupFuncWithMetricExpr(ec *EvalConfig, funcName string, rf rollupFunc
 	keepMetricNames := getKeepMetricNames(expr)
 	var tss []*timeseries
 	if iafc != nil {
-		tss, err = evalRollupWithIncrementalAggregate(funcName, keepMetricNames, iafc, rss, rcs, preFunc, sharedTimestamps)
+		tss, err = evalRollupWithIncrementalAggregate(qt, funcName, keepMetricNames, iafc, rss, rcs, preFunc, sharedTimestamps)
 	} else {
-		tss, err = evalRollupNoIncrementalAggregate(funcName, keepMetricNames, rss, rcs, preFunc, sharedTimestamps)
+		tss, err = evalRollupNoIncrementalAggregate(qt, funcName, keepMetricNames, rss, rcs, preFunc, sharedTimestamps)
 	}
 	if err != nil {
 		return nil, err
 	}
 	tss = mergeTimeseries(tssCached, tss, start, ec)
-	rollupResultCacheV.Put(ec, expr, window, tss)
+	rollupResultCacheV.Put(qt, ec, expr, window, tss)
 	return tss, nil
 }
 
@@ -910,23 +986,29 @@ func getRollupMemoryLimiter() *memoryLimiter {
 	return &rollupMemoryLimiter
 }
 
-func evalRollupWithIncrementalAggregate(funcName string, keepMetricNames bool, iafc *incrementalAggrFuncContext, rss *netstorage.Results, rcs []*rollupConfig,
+func evalRollupWithIncrementalAggregate(qt *querytracer.Tracer, funcName string, keepMetricNames bool,
+	iafc *incrementalAggrFuncContext, rss *netstorage.Results, rcs []*rollupConfig,
 	preFunc func(values []float64, timestamps []int64), sharedTimestamps []int64) ([]*timeseries, error) {
-	err := rss.RunParallel(func(rs *netstorage.Result, workerID uint) error {
+	qt = qt.NewChild("rollup %s() with incremental aggregation %s() over %d series; rollupConfigs=%s", funcName, iafc.ae.Name, rss.Len(), rcs)
+	defer qt.Done()
+	var samplesScannedTotal uint64
+	err := rss.RunParallel(qt, func(rs *netstorage.Result, workerID uint) error {
 		rs.Values, rs.Timestamps = dropStaleNaNs(funcName, rs.Values, rs.Timestamps)
 		preFunc(rs.Values, rs.Timestamps)
 		ts := getTimeseries()
 		defer putTimeseries(ts)
 		for _, rc := range rcs {
 			if tsm := newTimeseriesMap(funcName, keepMetricNames, sharedTimestamps, &rs.MetricName); tsm != nil {
-				rc.DoTimeseriesMap(tsm, rs.Values, rs.Timestamps)
+				samplesScanned := rc.DoTimeseriesMap(tsm, rs.Values, rs.Timestamps)
 				for _, ts := range tsm.m {
 					iafc.updateTimeseries(ts, workerID)
 				}
+				atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 				continue
 			}
 			ts.Reset()
-			doRollupForTimeseries(funcName, keepMetricNames, rc, ts, &rs.MetricName, rs.Values, rs.Timestamps, sharedTimestamps)
+			samplesScanned := doRollupForTimeseries(funcName, keepMetricNames, rc, ts, &rs.MetricName, rs.Values, rs.Timestamps, sharedTimestamps)
+			atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 			iafc.updateTimeseries(ts, workerID)
 
 			// ts.Timestamps points to sharedTimestamps. Zero it, so it can be re-used.
@@ -939,26 +1021,33 @@ func evalRollupWithIncrementalAggregate(funcName string, keepMetricNames bool, i
 		return nil, err
 	}
 	tss := iafc.finalizeTimeseries()
+	rowsScannedPerQuery.Update(float64(samplesScannedTotal))
+	qt.Printf("series after aggregation with %s(): %d; samplesScanned=%d", iafc.ae.Name, len(tss), samplesScannedTotal)
 	return tss, nil
 }
 
-func evalRollupNoIncrementalAggregate(funcName string, keepMetricNames bool, rss *netstorage.Results, rcs []*rollupConfig,
+func evalRollupNoIncrementalAggregate(qt *querytracer.Tracer, funcName string, keepMetricNames bool, rss *netstorage.Results, rcs []*rollupConfig,
 	preFunc func(values []float64, timestamps []int64), sharedTimestamps []int64) ([]*timeseries, error) {
+	qt = qt.NewChild("rollup %s() over %d series; rollupConfigs=%s", funcName, rss.Len(), rcs)
+	defer qt.Done()
 	tss := make([]*timeseries, 0, rss.Len()*len(rcs))
 	var tssLock sync.Mutex
-	err := rss.RunParallel(func(rs *netstorage.Result, workerID uint) error {
+	var samplesScannedTotal uint64
+	err := rss.RunParallel(qt, func(rs *netstorage.Result, workerID uint) error {
 		rs.Values, rs.Timestamps = dropStaleNaNs(funcName, rs.Values, rs.Timestamps)
 		preFunc(rs.Values, rs.Timestamps)
 		for _, rc := range rcs {
 			if tsm := newTimeseriesMap(funcName, keepMetricNames, sharedTimestamps, &rs.MetricName); tsm != nil {
-				rc.DoTimeseriesMap(tsm, rs.Values, rs.Timestamps)
+				samplesScanned := rc.DoTimeseriesMap(tsm, rs.Values, rs.Timestamps)
+				atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 				tssLock.Lock()
 				tss = tsm.AppendTimeseriesTo(tss)
 				tssLock.Unlock()
 				continue
 			}
 			var ts timeseries
-			doRollupForTimeseries(funcName, keepMetricNames, rc, &ts, &rs.MetricName, rs.Values, rs.Timestamps, sharedTimestamps)
+			samplesScanned := doRollupForTimeseries(funcName, keepMetricNames, rc, &ts, &rs.MetricName, rs.Values, rs.Timestamps, sharedTimestamps)
+			atomic.AddUint64(&samplesScannedTotal, samplesScanned)
 			tssLock.Lock()
 			tss = append(tss, &ts)
 			tssLock.Unlock()
@@ -968,11 +1057,13 @@ func evalRollupNoIncrementalAggregate(funcName string, keepMetricNames bool, rss
 	if err != nil {
 		return nil, err
 	}
+	rowsScannedPerQuery.Update(float64(samplesScannedTotal))
+	qt.Printf("samplesScanned=%d", samplesScannedTotal)
 	return tss, nil
 }
 
 func doRollupForTimeseries(funcName string, keepMetricNames bool, rc *rollupConfig, tsDst *timeseries, mnSrc *storage.MetricName,
-	valuesSrc []float64, timestampsSrc []int64, sharedTimestamps []int64) {
+	valuesSrc []float64, timestampsSrc []int64, sharedTimestamps []int64) uint64 {
 	tsDst.MetricName.CopyFrom(mnSrc)
 	if len(rc.TagValue) > 0 {
 		tsDst.MetricName.AddTag("rollup", rc.TagValue)
@@ -980,9 +1071,11 @@ func doRollupForTimeseries(funcName string, keepMetricNames bool, rc *rollupConf
 	if !keepMetricNames && !rollupFuncsKeepMetricName[funcName] {
 		tsDst.MetricName.ResetMetricGroup()
 	}
-	tsDst.Values = rc.Do(tsDst.Values[:0], valuesSrc, timestampsSrc)
+	var samplesScanned uint64
+	tsDst.Values, samplesScanned = rc.Do(tsDst.Values[:0], valuesSrc, timestampsSrc)
 	tsDst.Timestamps = sharedTimestamps
 	tsDst.denyReuse = true
+	return samplesScanned
 }
 
 var bbPool bytesutil.ByteBufferPool
