@@ -22,11 +22,15 @@ type SDConfig struct {
 
 	// Use role() function for accessing the Role field
 	Role string `yaml:"role"`
+	// The filepath to kube config.
+	// If defined any cluster connection information from HTTPClientConfig is ignored.
+	KubeConfigFile string `yaml:"kubeconfig_file"`
 
 	HTTPClientConfig promauth.HTTPClientConfig `yaml:",inline"`
 	ProxyURL         *proxy.URL                `yaml:"proxy_url,omitempty"`
 	Namespaces       Namespaces                `yaml:"namespaces,omitempty"`
 	Selectors        []Selector                `yaml:"selectors,omitempty"`
+	AttachMetadata   AttachMetadata            `yaml:"attach_metadata,omitempty"`
 
 	cfg      *apiConfig
 	startErr error
@@ -41,6 +45,13 @@ func (sdc *SDConfig) role() string {
 	return sdc.Role
 }
 
+// AttachMetadata represents `attach_metadata` option at `kuberentes_sd_config`.
+//
+// See https://prometheus.io/docs/prometheus/latest/configuration/configuration/#kubernetes_sd_config
+type AttachMetadata struct {
+	Node bool `yaml:"node"`
+}
+
 // Namespaces represents namespaces for SDConfig
 type Namespaces struct {
 	OwnNamespace bool     `yaml:"own_namespace"`
