@@ -245,7 +245,7 @@ func (sw *scrapeWork) loadLastScrape() string {
 }
 
 func (sw *scrapeWork) storeLastScrape(lastScrape []byte) {
-	mustCompress := minResponseSizeForStreamParse.N > 0 && len(lastScrape) >= minResponseSizeForStreamParse.N
+	mustCompress := minResponseSizeForStreamParse.N > 0 && len(lastScrape) >= minResponseSizeForStreamParse.IntN()
 	if mustCompress {
 		sw.lastScrapeCompressed = encoding.CompressZSTDLevel(sw.lastScrapeCompressed[:0], lastScrape, 1)
 		sw.lastScrape = nil
@@ -400,7 +400,7 @@ func (sw *scrapeWork) mustSwitchToStreamParseMode(responseSize int) bool {
 	if minResponseSizeForStreamParse.N <= 0 {
 		return false
 	}
-	return sw.Config.canSwitchToStreamParseMode() && responseSize >= minResponseSizeForStreamParse.N
+	return sw.Config.canSwitchToStreamParseMode() && responseSize >= minResponseSizeForStreamParse.IntN()
 }
 
 // getTargetResponse() fetches response from sw target in the same way as when scraping the target.
@@ -745,7 +745,7 @@ func (sw *scrapeWork) sendStaleSeries(lastScrape, currScrape string, timestamp i
 	}
 	wc := &writeRequestCtx{}
 	if bodyString != "" {
-		wc.rows.Unmarshal(bodyString)
+		wc.rows.UnmarshalWithErrLogger(bodyString, sw.logError)
 		srcRows := wc.rows.Rows
 		for i := range srcRows {
 			sw.addRowToTimeseries(wc, &srcRows[i], timestamp, true)
