@@ -3,6 +3,7 @@ package storage
 import (
 	"bytes"
 	"fmt"
+	"runtime"
 	"sort"
 	"strconv"
 	"strings"
@@ -384,6 +385,14 @@ func (mn *MetricName) Marshal(dst []byte) []byte {
 	return dst
 }
 
+// UnmarshalString unmarshals mn from s
+func (mn *MetricName) UnmarshalString(s string) error {
+	b := bytesutil.ToUnsafeBytes(s)
+	err := mn.Unmarshal(b)
+	runtime.KeepAlive(s)
+	return err
+}
+
 // Unmarshal unmarshals mn from src.
 func (mn *MetricName) Unmarshal(src []byte) error {
 	// Unmarshal MetricGroup.
