@@ -17,7 +17,8 @@ func TestTableOpenClose(t *testing.T) {
 	}()
 
 	// Create a new table
-	tb, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs)
+	var isReadOnly uint32
+	tb, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs, &isReadOnly)
 	if err != nil {
 		t.Fatalf("cannot create new table: %s", err)
 	}
@@ -27,7 +28,7 @@ func TestTableOpenClose(t *testing.T) {
 
 	// Re-open created table multiple times.
 	for i := 0; i < 10; i++ {
-		tb, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs)
+		tb, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs, &isReadOnly)
 		if err != nil {
 			t.Fatalf("cannot open created table: %s", err)
 		}
@@ -43,14 +44,15 @@ func TestTableOpenMultipleTimes(t *testing.T) {
 		_ = os.RemoveAll(path)
 	}()
 
-	tb1, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs)
+	var isReadOnly uint32
+	tb1, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs, &isReadOnly)
 	if err != nil {
 		t.Fatalf("cannot open table the first time: %s", err)
 	}
 	defer tb1.MustClose()
 
 	for i := 0; i < 10; i++ {
-		tb2, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs)
+		tb2, err := openTable(path, nilGetDeletedMetricIDs, retentionMsecs, &isReadOnly)
 		if err == nil {
 			tb2.MustClose()
 			t.Fatalf("expecting non-nil error when opening already opened table")
